#ifndef BT_UNWIND_H
#define BT_UNWIND_H

#include <sys/types.h>

#include "dwarf.h"
#include "error.h"
#include "register.h"

struct bt_process;
struct bt_thread;
struct bt_unwind;
struct bt_unwind_cursor;

enum bt_unwind_type {
	BT_UNWIND_BT = 0
};

enum bt_unwind_result {
	BT_UNWIND_RESULT_SUCCESS,	/* Frame unwound successfully */
	BT_UNWIND_RESULT_DONE,		/* Stack complete; no more frames to unwind. */
	BT_UNWIND_RESULT_FAILURE	/* Error occurred unwinding frame */
};

enum bt_unwind_frame_type {
	BT_UNWIND_FRAME_NORMAL,
	BT_UNWIND_FRAME_SIGNAL,
	BT_UNWIND_FRAME_TRAP
};

/*
 * Initialize the unwind subsystem.
 * Returns true on success and false otherwise.
 */
bool bt_unwind_init(void);

/*
 * Initialize an unwinder for a live process.
 * Returns true on success and false otherwise.
 */
typedef bool bt_unwind_init_live_fn_t(struct bt_unwind *,
    struct bt_process *, pid_t, bt_error_t *);

/*
 * Initialize an unwinder for a userspace core dump.
 * Returns true on success and false otherwise.
 */
typedef bool bt_unwind_init_core_fn_t(struct bt_unwind *,
    struct bt_process *, const char *, bt_error_t *);

/*
 * Initialize an unwinder for a kernel core dump.
 * Returns true on success and false otherwise.
 */
typedef bool bt_unwind_init_kernel_core_fn_t(struct bt_unwind *,
    struct bt_process *, const char *, bt_error_t *);

/*
 * Deinitialize an unwinder.
 */
typedef void bt_unwind_deinit_fn_t(struct bt_unwind *, struct bt_process *);

/*
 * Initialize an unwinding cursor.
 */
typedef bool bt_unwind_cursor_init_fn_t(struct bt_unwind_cursor *,
    bt_error_t *);

/*
 * Deinitialize an unwinding cursor.
 */
typedef enum bt_unwind_result bt_unwind_fn_t(struct bt_unwind_cursor *,
     bt_error_t *);

struct bt_unwind_adapter {
	bt_unwind_init_live_fn_t *init_live;
	bt_unwind_init_core_fn_t *init_core;
	bt_unwind_init_kernel_core_fn_t *init_kernel_core;
	bt_unwind_deinit_fn_t *deinit;
	bt_unwind_cursor_init_fn_t *cursor_init;
	bt_unwind_fn_t *unwind;
};

struct bt_unwind {
	enum bt_unwind_type type;
	struct bt_unwind_adapter adapter;
};
typedef struct bt_unwind bt_unwind_t;

struct bt_unwind_file;

/*
 * Returns a handle for unwinding frames from a given file.
 * Returns NULL if this adapters cannot read unwinding information from the
 * provided file.
 */
typedef void *bt_unwind_file_init_fn_t(struct bt_bf *);

/*
 * Unwind the provided cursor.  See bt_unwind_next for details about unwinding.
 */
typedef enum bt_unwind_result bt_unwind_file_unwind_fn_t(void *,
    struct bt_unwind_cursor *, bt_error_t *);

/*
 * Clean up any resources allocated when using the unwinder handle.
 * The argument is a handle returned by the above initialization function.
 */
typedef void bt_unwind_file_deinit_fn_t(void *);

struct bt_unwind_file_adapter {
	bt_unwind_file_init_fn_t *init;
	bt_unwind_file_unwind_fn_t *unwind;
	bt_unwind_file_deinit_fn_t *deinit;
};

/*
 * Register an adapter for extracting unwind information from object files.
 * Returns true on success and false otherwise.
 */
bool bt_unwind_file_adapter_register(const struct bt_unwind_file_adapter *);

/*
 * Representation of unwinding information for a specific object
 */
struct bt_unwind_file {
	void *context;
	struct bt_unwind_file_adapter adapter;
};
typedef struct bt_unwind_file bt_unwind_file_t;

static inline bool
bt_unwind_file_valid(const struct bt_unwind_file *unwind)
{

	return unwind->context != NULL;
}

/* 
 * Initialize an unwinder for a file using the appropriate debug information.
 * Returns true on success and false otherwise.
 */
bool bt_unwind_file_init(bt_unwind_file_t *, struct bt_bf *);

enum bt_unwind_result bt_unwind_file_unwind(bt_unwind_file_t *,
    struct bt_unwind_cursor *, bt_error_t *);

/*
 * Deinitialize a file's unwind handle.
 */
void bt_unwind_file_deinit(bt_unwind_file_t *);

#define BT_UNWIND_SIGNAL		(1U << 0)
#define BT_UNWIND_REGISTERS		(1U << 1)
#define BT_UNWIND_FRAME			(1U << 2)
#define BT_UNWIND_THREAD		(1U << 3)
#define BT_UNWIND_SIGNAL_CHILD		(1U << 4)

/*
 * This frame has signal information despite the absence of a signal trampoline.
 * This is most commonly seen in core files where an unhandled signal causes
 * a process to dump core.
 */
#define BT_UNWIND_SIGNAL_ARTIFICIAL	(1U << 5)

/*
 * This frame was unwound by applying architecture-specific heuristics instead
 * of actual debug information so its contents may not be accurate.
 */
#define BT_UNWIND_HEURISTIC		(1U << 6)

/*
 * This frame effectively raised a signal because either it directly raised a
 * signal or one of its inlined frames did.
 */
#define BT_UNWIND_SIGNAL_CONTAINED (BT_UNWIND_SIGNAL | BT_UNWIND_SIGNAL_CHILD)

struct bt_unwind_cursor {
	struct bt_unwind_cursor *caller;
	struct bt_unwind_cursor *callee;
	unsigned int flags;
	struct bt_process *process;
	struct bt_thread *thread;
	struct bt_register_set registers;
	unsigned long long fb;
	unsigned long long cfa;
	enum bt_unwind_frame_type frame_type;
	enum bt_unwind_type type;
};
typedef struct bt_unwind_cursor bt_unwind_cursor_t;

static inline enum bt_unwind_frame_type
bt_unwind_cursor_frame_type(const struct bt_unwind_cursor *cursor)
{

	return cursor->frame_type;
}

static inline bool
bt_unwind_cursor_has(const struct bt_unwind_cursor *cursor, unsigned int flags)
{

	return cursor->flags & flags;
}

/*
 * Returns an address that is guaranteed to be within the address limits of a
 * frame's function.  If a function foo calls function bar and that call
 * instruction is the last instruction in foo, then the instruction pointer for
 * foo's frame will be the return address for that call, which is past the end
 * of foo's range.  Therefore for any frame which has called another frame, we
 * should subtract one from the unwound instruction pointer to ensure that we
 * end up with an address within the caller's range.
 */
static inline unsigned long long
bt_unwind_cursor_effective_pc(struct bt_unwind_cursor *cursor)
{
	bool adjust = (bt_unwind_cursor_has(cursor,
	    BT_UNWIND_SIGNAL_CONTAINED) == false) & (cursor->callee != NULL);

	return cursor->registers.ip - adjust;
}

/*
 * Initialize an unwinding cursor for a thread.
 * Returns true on success and false otherwise.
 */
bool bt_unwind_cursor_init_thread(bt_unwind_cursor_t *,
    struct bt_thread *, bt_error_t *);

/*
 * Initialize an unwinding cursor for a process.
 * Returns true on success and false otherwise.
 */
void bt_unwind_cursor_init_process(bt_unwind_cursor_t *, struct bt_process *);

/*
 * Initialize an unwinding cursor for a frame.
 * Returns true on success and false otherwise.
 */
void bt_unwind_cursor_frame(bt_unwind_cursor_t *, unsigned long long,
    unsigned long long);

/*
 * Remove any frame context from an unwinding cursor.
 */
void bt_unwind_cursor_frame_clear(bt_unwind_cursor_t *);

/*
 * Unwind the provided cursor.  That is, given its current frame state, update
 * it to contain the state of the calling frame.
 * The first argument is a handle returned by the above initialization function.
 * Resulting cursor contents are valid iff the return value is
 * BT_UNWIND_RESULT_SUCCESS.
 * A return value of BT_UNWIND_RESULT_DONE indicates that the provided cursor
 * represented the top-most frame in its stack.
 * A return value of BT_UNWIND_RESULT_FAILURE indicates that an error occurred
 * unwinding the provided cursor.
 */
enum bt_unwind_result bt_unwind_next(bt_unwind_cursor_t *, bt_error_t *);

/*
 * Default implementation of libunwind callback for finding the debug file for
 * an address.
 */
bool bt_unwind_get_elf_image(struct bt_process *, unsigned long long,
    unsigned long *, unsigned long *, char *, size_t, bt_error_t *);

/*
 * Initializes unwind context for live process.
 */
bool bt_unwind_init_live(bt_unwind_t *, struct bt_process *, pid_t,
    bt_error_t *);

/*
 * Initializes unwind context for user-space core dump.
 */
bool bt_unwind_init_core(bt_unwind_t *, struct bt_process *,
    const char *, bt_error_t *);

/*
 * Initializes unwind context for kernel core dump.
 */
bool bt_unwind_init_core_kernel(bt_unwind_t *, struct bt_process *,
    const char *, bt_error_t *);

/*
 * Destroys unwind context.
 */
void bt_unwind_deinit(bt_unwind_t *, struct bt_process *);

#endif /* BT_UNWIND_H */
