#ifndef _BT_REGISTER_H
#define _BT_REGISTER_H

#include <ck_cc.h>
#include <string.h>
#include <limits.h>

#include "bitmap.h"

#if defined(__sun__)
#include <sys/regset.h>
#endif /* __sun__ */

#define BT_REGISTER_MASK(R) (1UL << (R))

#if defined(__x86_64__)
/*
 * This is DWARF register mapping according to:
 * http://www.x86-64.org/documentation/abi.pdf
 */
enum bt_register_md {
	BT_REGISTER_RAX = 0,
	BT_REGISTER_RDX,
	BT_REGISTER_RCX,
	BT_REGISTER_RBX,
	BT_REGISTER_RSI,
	BT_REGISTER_RDI,
	BT_REGISTER_RBP,
	BT_REGISTER_RSP,
	BT_REGISTER_R8,
	BT_REGISTER_R9,
	BT_REGISTER_R10,
	BT_REGISTER_R11,
	BT_REGISTER_R12,
	BT_REGISTER_R13,
	BT_REGISTER_R14,
	BT_REGISTER_R15,
	BT_REGISTER_RIP,
	BT_REGISTER_CS,
	BT_REGISTER_SS,
	BT_REGISTER_DS,
	BT_REGISTER_ES,
	BT_REGISTER_FS,
	BT_REGISTER_GS,
	BT_REGISTER_EFLAGS,
	BT_REGISTER_COUNT
};

#define BT_REGISTER_GP_WIDTH	8
#define BT_REGISTER_IP BT_REGISTER_RIP
#define BT_REGISTER_SP BT_REGISTER_RSP

static inline unsigned long
bt_register_mask(unsigned long r)
{

	return (r < BT_REGISTER_COUNT) ? (1UL << r) : 0;
}

static CK_CC_UNUSED const char *
bt_register_name(int n)
{
	static const char *lut[] = {
		[BT_REGISTER_RAX] = "rax",
		[BT_REGISTER_RDX] = "rdx",
		[BT_REGISTER_RCX] = "rcx",
		[BT_REGISTER_RBX] = "rbx",
		[BT_REGISTER_RSI] = "rsi",
		[BT_REGISTER_RDI] = "rdi",
		[BT_REGISTER_RBP] = "rbp",
		[BT_REGISTER_RSP] = "rsp",
		[BT_REGISTER_R8] = "r8",
		[BT_REGISTER_R9] = "r9",
		[BT_REGISTER_R10] = "r10",
		[BT_REGISTER_R11] = "r11",
		[BT_REGISTER_R12] = "r12",
		[BT_REGISTER_R13] = "r13",
		[BT_REGISTER_R14] = "r14",
		[BT_REGISTER_R15] = "r15",
		[BT_REGISTER_RIP] = "rip",
		[BT_REGISTER_CS] = "cs",
		[BT_REGISTER_SS] = "ss",
		[BT_REGISTER_DS] = "ds",
		[BT_REGISTER_ES] = "es",
		[BT_REGISTER_FS] = "fs",
		[BT_REGISTER_GS] = "gs",
		[BT_REGISTER_EFLAGS] = "eflags",
	};

	if (n >= BT_REGISTER_COUNT || n < 0)
		return "unknown";

	return lut[n];
}

typedef unsigned long long bt_register_gp_t;

#if defined(__linux__)
#include <sys/user.h>
typedef struct user_regs_struct bt_register_md_gp_t;
typedef struct user_fpregs_struct bt_register_md_fp_t;
#elif defined(__FreeBSD__)
#include <sys/ptrace.h>
typedef struct reg bt_register_md_gp_t;
typedef struct fpreg bt_register_md_fp_t;
#elif defined(__sun__)
typedef gregset_t bt_register_md_gp_t;
typedef fpregset_t bt_register_md_fp_t;
#endif /* __sun__ */
#endif /* __x86_64__ */

struct bt_register_set {
	unsigned long fp;
	unsigned long sp;
	unsigned long ip;

	struct {
		/* Fields set if field available. */
		BT_BITMAP_INSTANCE(BT_REGISTER_COUNT) a_gp;
		bt_register_gp_t gp[BT_REGISTER_COUNT];
		bt_register_md_fp_t fp;
	} md;
};

static inline bt_register_gp_t *
bt_register_address_gp(struct bt_register_set *r, enum bt_register_md md)
{

	if (md >= BT_REGISTER_COUNT)
		return NULL;

	return &r->md.gp[md];
}

static inline void
bt_register_inherit(struct bt_register_set *r, const struct bt_register_set *s)
{

	BT_BITMAP_UNION(&r->md.a_gp, &s->md.a_gp);
	return;
}

static inline void
bt_register_reset(struct bt_register_set *r)
{

	BT_BITMAP_INIT(&r->md.a_gp, BT_REGISTER_COUNT, false);
	return;
}

static inline void
bt_register_set_copy(struct bt_register_set *d, struct bt_register_set *s)
{

	memcpy(d, s, sizeof *d);
	return;
}

static inline bool
bt_register_gp_set(struct bt_register_set *r, enum bt_register_md md,
    bt_register_gp_t value)
{
	bt_register_gp_t *gp;

	gp = bt_register_address_gp(r, md);
	if (gp == NULL)
		return false;

	*gp = value;
	BT_BITMAP_SET(&r->md.a_gp, md);
	return true;
}

static inline void
bt_register_gp_invalidate(struct bt_register_set *r, enum bt_register_md md)
{

	BT_BITMAP_RESET(&r->md.a_gp, md);
	return;
}

static inline bool
bt_register_gp_get(struct bt_register_set *s, enum bt_register_md md,
    bt_register_gp_t *value)
{
	bt_register_gp_t *r;

	r = bt_register_address_gp(s, md);
	if (r == NULL)
		return false;

	*value = *r;
	return BT_BITMAP_TEST(&s->md.a_gp, md);
}

/*
 * Copy the value of register src to register dst.  If src's value is unknown
 * then dst will be invalidated as well.
 */
static inline void
bt_register_gp_copy(struct bt_register_set *s, enum bt_register_md dst,
    enum bt_register_md src)
{
	bt_register_gp_t value;

	if (bt_register_gp_get(s, src, &value) == true) {
		bt_register_gp_set(s, dst, value);
	} else {
		bt_register_gp_invalidate(s, dst);
	}

	return;
}

/*
 * Maps DWARF register slots to libbt register slots.
 */
static inline enum bt_register_md
bt_register_regnum_dw(int r)
{

	switch (r) {
	case 0:
		return BT_REGISTER_RAX;
	case 1:
		return BT_REGISTER_RDX;
	case 2:
		return BT_REGISTER_RCX;
	case 3:
		return BT_REGISTER_RBX;
	case 4:
		return BT_REGISTER_RSI;
	case 5:
		return BT_REGISTER_RDI;
	case 6:
		return BT_REGISTER_RBP;
	case 7:
		return BT_REGISTER_RSP;
	case 8:
		return BT_REGISTER_R8;
	case 9:
		return BT_REGISTER_R9;
	case 10:
		return BT_REGISTER_R10;
	case 11:
		return BT_REGISTER_R11;
	case 12:
		return BT_REGISTER_R12;
	case 13:
		return BT_REGISTER_R13;
	case 14:
		return BT_REGISTER_R14;
	case 15:
		return BT_REGISTER_R15;
	case 16:
		return BT_REGISTER_RIP;
	case 49:
		return BT_REGISTER_EFLAGS;
	case 50:
		return BT_REGISTER_ES;
	case 51:
		return BT_REGISTER_CS;
	case 52:
		return BT_REGISTER_SS;
	case 53:
		return BT_REGISTER_DS;
	case 54:
		return BT_REGISTER_FS;
	case 55:
		return BT_REGISTER_GS;
	case 58:
		return BT_REGISTER_FS;
	default:
		break;
	}

	return BT_REGISTER_COUNT;
}

static inline bt_register_gp_t *
bt_register_address_gp_dw(struct bt_register_set *r, int rn)
{
	enum bt_register_md md = bt_register_regnum_dw(rn);

	if (BT_BITMAP_TEST(&r->md.a_gp, md) == false)
		return NULL;

	return bt_register_address_gp(r, md);
}

/*
 * Initializes structure using target-specific structures.
 */
static inline void
bt_register_set_init(struct bt_register_set *r,
    void *pun, bt_register_md_fp_t *fp)
{
	union {
		bt_register_md_gp_t *gp;
		void *pun;
	} gp;

	gp.pun = pun;

#ifdef __x86_64__
#ifdef __linux__
	r->md.gp[BT_REGISTER_RAX] = gp.gp->rax;
	r->md.gp[BT_REGISTER_RDX] = gp.gp->rdx;
	r->md.gp[BT_REGISTER_RCX] = gp.gp->rcx;
	r->md.gp[BT_REGISTER_RBX] = gp.gp->rbx;
	r->md.gp[BT_REGISTER_RSI] = gp.gp->rsi;
	r->md.gp[BT_REGISTER_RDI] = gp.gp->rdi;
	r->md.gp[BT_REGISTER_RBP] = gp.gp->rbp;
	r->md.gp[BT_REGISTER_RSI] = gp.gp->rsi;
	r->md.gp[BT_REGISTER_RDI] = gp.gp->rdi;
	r->fp = r->md.gp[BT_REGISTER_RBP] = gp.gp->rbp;
	r->sp = r->md.gp[BT_REGISTER_RSP] = gp.gp->rsp;
	r->md.gp[BT_REGISTER_R8] = gp.gp->r8;
	r->md.gp[BT_REGISTER_R9] = gp.gp->r9;
	r->md.gp[BT_REGISTER_R10] = gp.gp->r10;
	r->md.gp[BT_REGISTER_R11] = gp.gp->r11;
	r->md.gp[BT_REGISTER_R12] = gp.gp->r12;
	r->md.gp[BT_REGISTER_R13] = gp.gp->r13;
	r->md.gp[BT_REGISTER_R14] = gp.gp->r14;
	r->md.gp[BT_REGISTER_R15] = gp.gp->r15;
	r->ip = r->md.gp[BT_REGISTER_RIP] = gp.gp->rip;
	r->md.gp[BT_REGISTER_CS] = gp.gp->cs;
	r->md.gp[BT_REGISTER_SS] = gp.gp->ss;
	r->md.gp[BT_REGISTER_DS] = gp.gp->ds;
	r->md.gp[BT_REGISTER_ES] = gp.gp->es;
	r->md.gp[BT_REGISTER_FS] = gp.gp->fs;
	r->md.gp[BT_REGISTER_GS] = gp.gp->gs;
	r->md.gp[BT_REGISTER_EFLAGS] = gp.gp->eflags;
	BT_BITMAP_INIT(&r->md.a_gp, BT_REGISTER_COUNT, true);

	if (fp != NULL)
		r->md.fp = *fp;
#elif defined(__FreeBSD__)
	r->md.gp[BT_REGISTER_RAX] = gp.gp->r_rax;
	r->md.gp[BT_REGISTER_RDX] = gp.gp->r_rdx;
	r->md.gp[BT_REGISTER_RCX] = gp.gp->r_rcx;
	r->md.gp[BT_REGISTER_RBX] = gp.gp->r_rbx;
	r->md.gp[BT_REGISTER_RSI] = gp.gp->r_rsi;
	r->md.gp[BT_REGISTER_RDI] = gp.gp->r_rdi;
	r->md.gp[BT_REGISTER_RBP] = gp.gp->r_rbp;
	r->md.gp[BT_REGISTER_RSI] = gp.gp->r_rsi;
	r->md.gp[BT_REGISTER_RDI] = gp.gp->r_rdi;
	r->fp = r->md.gp[BT_REGISTER_RBP] = gp.gp->r_rbp;
	r->sp = r->md.gp[BT_REGISTER_RSP] = gp.gp->r_rsp;
	r->md.gp[BT_REGISTER_R8] = gp.gp->r_r8;
	r->md.gp[BT_REGISTER_R9] = gp.gp->r_r9;
	r->md.gp[BT_REGISTER_R10] = gp.gp->r_r10;
	r->md.gp[BT_REGISTER_R11] = gp.gp->r_r11;
	r->md.gp[BT_REGISTER_R12] = gp.gp->r_r12;
	r->md.gp[BT_REGISTER_R13] = gp.gp->r_r13;
	r->md.gp[BT_REGISTER_R14] = gp.gp->r_r14;
	r->md.gp[BT_REGISTER_R15] = gp.gp->r_r15;
	r->ip = r->md.gp[BT_REGISTER_RIP] = gp.gp->r_rip;

	r->md.gp[BT_REGISTER_CS] = gp.gp->r_cs;
	r->md.gp[BT_REGISTER_SS] = gp.gp->r_ss;
	r->md.gp[BT_REGISTER_DS] = gp.gp->r_ds;
	r->md.gp[BT_REGISTER_ES] = gp.gp->r_es;
	r->md.gp[BT_REGISTER_FS] = gp.gp->r_fs;
	r->md.gp[BT_REGISTER_GS] = gp.gp->r_gs;
	r->md.gp[BT_REGISTER_EFLAGS] = gp.gp->r_rflags;
	BT_BITMAP_INIT(&r->md.a_gp, BT_REGISTER_COUNT, true);

	if (fp != NULL)
		r->md.fp = *fp;
#elif defined(__sun__)
	r->md.gp[BT_REGISTER_RAX]  = (*gp.gp)[REG_RAX];
	r->md.gp[BT_REGISTER_RDX]  = (*gp.gp)[REG_RDX];
	r->md.gp[BT_REGISTER_RCX]  = (*gp.gp)[REG_RCX];
	r->md.gp[BT_REGISTER_RBX]  = (*gp.gp)[REG_RBX];
	r->md.gp[BT_REGISTER_RSI]  = (*gp.gp)[REG_RSI];
	r->md.gp[BT_REGISTER_RDI]  = (*gp.gp)[REG_RDI];
	r->md.gp[BT_REGISTER_RBP]  = (*gp.gp)[REG_RBP];
	r->md.gp[BT_REGISTER_RSI]  = (*gp.gp)[REG_RSI];
	r->md.gp[BT_REGISTER_RDI]  = (*gp.gp)[REG_RDI];
	r->fp = r->md.gp[BT_REGISTER_RBP]  = (*gp.gp)[REG_RBP];
	r->sp = r->md.gp[BT_REGISTER_RSP]  = (*gp.gp)[REG_RSP];
	r->md.gp[BT_REGISTER_R8]  = (*gp.gp)[REG_R8];
	r->md.gp[BT_REGISTER_R9]  = (*gp.gp)[REG_R9];
	r->md.gp[BT_REGISTER_R10]  = (*gp.gp)[REG_R10];
	r->md.gp[BT_REGISTER_R11]  = (*gp.gp)[REG_R11];
	r->md.gp[BT_REGISTER_R12]  = (*gp.gp)[REG_R12];
	r->md.gp[BT_REGISTER_R13]  = (*gp.gp)[REG_R13];
	r->md.gp[BT_REGISTER_R14]  = (*gp.gp)[REG_R14];
	r->md.gp[BT_REGISTER_R15]  = (*gp.gp)[REG_R15];
	r->ip = r->md.gp[BT_REGISTER_RIP]  = (*gp.gp)[REG_RIP];
	BT_BITMAP_INIT(&r->md.a_gp, BT_REGISTER_COUNT, true);

	if (fp != NULL)
		r->md.fp = *fp;
#endif /* __sun__ */
#endif /* __x86_64__ */
	return;
}

typedef unsigned long long bt_register_mask_t;

#define BT_NO_REGS_USED 0

#if defined(__x86_64__)
static const bt_register_mask_t bt_scratch_reg_mask =
	BT_REGISTER_MASK(BT_REGISTER_RAX) |
	BT_REGISTER_MASK(BT_REGISTER_RDX) |
	BT_REGISTER_MASK(BT_REGISTER_RCX) |
	BT_REGISTER_MASK(BT_REGISTER_RSI) |
	BT_REGISTER_MASK(BT_REGISTER_RDI) |
	BT_REGISTER_MASK(BT_REGISTER_R8)  |
	BT_REGISTER_MASK(BT_REGISTER_R9)  |
	BT_REGISTER_MASK(BT_REGISTER_R10) |
	BT_REGISTER_MASK(BT_REGISTER_R11);

static const bt_register_mask_t bt_register_callee_saved_mask =
	BT_REGISTER_MASK(BT_REGISTER_RBP) |
	BT_REGISTER_MASK(BT_REGISTER_RBX) |
	BT_REGISTER_MASK(BT_REGISTER_R12) |
	BT_REGISTER_MASK(BT_REGISTER_R13) |
	BT_REGISTER_MASK(BT_REGISTER_R14) |
	BT_REGISTER_MASK(BT_REGISTER_R15);
#endif /* __x86_64__ */

static inline void
bt_register_mask_init(bt_register_mask_t *reg_mask,
    bt_register_mask_t init_mask)
{

	*reg_mask = init_mask;
	return;
}

static inline void
bt_register_mask_set(bt_register_mask_t *reg_mask, unsigned long long reg)
{

	*reg_mask |= bt_register_mask(reg);
	return;
}

static inline bt_register_mask_t
bt_register_mask_get(bt_register_mask_t *reg_mask)
{

	return *reg_mask;
}

static inline bool
bt_register_mask_test(const bt_register_mask_t *mask, enum bt_register_md reg)
{

	return *mask & bt_register_mask(reg);
}

#endif /* _BT_REGISTER_H */
