#ifndef _BT_SO_H
#define _BT_SO_H

#include <ck_hs.h>
#include <stdbool.h>
#include "lut.h"
#include "array.h"

/*
 * BT_SO_STATIC may also mean that the binary was compiled
 * with no linking at all, which will (in the case of -nostdlib)
 * in some cases leave it devoid of dynamic information and thus
 * will appear as static.
 */
enum bt_so_linking {
	BT_SO_STATIC,
	BT_SO_DYNAMIC
};

/*
 * This file provides an interface for resolving paths to shared objects
 * required by executables.
 */

/*
 * Flags indiciating whether or not rpath or runpath were expanded. This
 * is important for us to know, because if they were expanded then they
 * are not pointing to dynstr + some_offset. But rather heap allocated
 * memory and must be freed at some point.
 */
#define BT_SO_F_RPATH_EXPANSION (1U << 0)
#define BT_SO_F_RUNPATH_EXPANSION (1U << 1)

struct bt_so_resolver {
	/*
	 * This holds the class of the executable program who is
	 * needing the shared libraries we are resolving to be linked.
	 */
	enum bt_bf_class linking_class;

	/*
	 * Is the process spawned from a program that is dynamically
	 * or statically linked?
	 */
	enum bt_so_linking linking_type;

	/*
	 * Mapping from basename to absolute path.
	 * Entries in this map take predecence over searching paths.
	 */
	ck_hs_t custom_paths;

	/*
	 * List of custom directories to search.  These directories will
	 * take precedence over any other paths that are normally searched.
	 */
	bt_lut_t custom_dirs;
	unsigned int n_custom_dir;

	char *rpath;
	char *runpath;

	unsigned int flag;

	/*
	 * The maximum number of dependency levels to search. If this value is
	 * 0, then an unlimited number of levels will be searched.
	 */
	unsigned int levels;
};

bool bt_so_resolver_init(struct bt_so_resolver *);
void bt_so_resolver_destroy(struct bt_so_resolver *);
bool bt_so_resolver_set_class(struct bt_so_resolver *, enum bt_bf_class);

bool bt_so_resolver_custom_dir_add(struct bt_so_resolver *, char *);
bool bt_so_resolver_custom_path_add(struct bt_so_resolver *, char *, char *);

/*
 * Resolves the basename of a shared object into the proper absolute path.
 * If the function returns true, the buffer pointed to by the first argument
 * will contain the absolute path of the resolved object.
 * This buffer must have at least PATH_MAX bytes available.
 */
bool bt_so_resolve(char *, struct bt_so_resolver *, char *);

/*
 * iterator entry for bt_so_needed_iterator
 */
struct bt_so_needed_entry {
	char *so_path;
};
typedef struct bt_so_needed_entry bt_so_needed_entry_t;

struct bt_so_needed_node;

/*
 * OK: iteration succeeded
 * DONE: iteration complete
 * ERROR: iteration had fatal error
 * EMPTY: iteration found no shared objects
 * FAILED: iteration failed to resolve a given object
 */
enum bt_so_iterator_result {
	BT_SO_ITER_OK,
	BT_SO_ITER_DONE,
	BT_SO_ITER_ERROR,
	BT_SO_ITER_FAILED
};

struct bt_so_needed_iterator {
	bt_bf_t **bf;
	size_t bf_count;
	struct bt_so_resolver *resolver;
	size_t bf_index;
	const char *obj_path;
	char *rpath;
	char *runpath;
	unsigned int depth;
	unsigned int level;
	SLIST_HEAD(bt_so_helper_list, bt_so_needed_node) helper_list;
	SLIST_HEAD(bt_so_yield_list, bt_so_needed_node) yield_list;
	struct bt_so_needed_node *yield;
	struct bt_so_needed_node *helper;
	ck_hs_t lookup;
};
typedef struct bt_so_needed_iterator bt_so_needed_iterator_t;

/*
 * Initializes iterator for enumerating needed shared objects transitively.
 * The 3rd argument is the number of bf objects in the array
 * The 4th argument is optional, if the caller doesn't supply a resolver
 * then one will be created.
 * The 5th argument (unsigned int) is the depth, specifying how many
 * levels deep to follow the needed objects. 0 means no limit
 */
bool bt_so_needed_iterator_init(bt_so_needed_iterator_t *, bt_bf_t **,
    size_t, struct bt_so_resolver *, unsigned int);

/*
 * Deinitialize iterator: free each path associated with each entry.
 */
void bt_so_needed_iterator_deinit(bt_so_needed_iterator_t *);

/*
 * Iterate over each shared library dependency by transitively following
 * the needed entries of each object.
 */
enum bt_so_iterator_result
bt_so_needed_iterator_next(bt_so_needed_iterator_t *,
    bt_so_needed_entry_t *, bt_error_t *);

/*
 * Returns the entry->so_path
 */
static inline char *
bt_so_needed_path(bt_so_needed_entry_t *entry)
{

	return entry->so_path;
}

/*
 * Entry type for the preloaded iterator
 */
struct bt_so_preloaded_entry {
	char *so_path;
};
typedef struct bt_so_preloaded_entry bt_so_preloaded_entry_t;

struct bt_so_preloaded_iterator {
	unsigned int index;
	struct bt_process *process;
	char *parse_string;
	char *tmp;
	struct bt_so_resolver resolver;
};
typedef struct bt_so_preloaded_iterator bt_so_preloaded_iterator_t;

/*
 * Initialize preloaded iterator.
 */
bool bt_so_preloaded_iterator_init(bt_so_preloaded_iterator_t *,
    struct bt_process *);

/*
 * Deinitialize the preloaded iterator.
 */
void bt_so_preloaded_iterator_deinit(bt_so_preloaded_iterator_t *);

/*
 * Iterate over each shared library path specified by LD_PRELOAD.
 */
enum bt_so_iterator_result
bt_so_preloaded_iterator_next(bt_so_preloaded_iterator_t *,
    bt_so_preloaded_entry_t *, bt_error_t *);

/*
 * Returns the entry->so_path
 */
char *bt_so_preloaded_path(bt_so_preloaded_entry_t *);

#endif /* _BT_SO_H */
